#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <usb.h>
#include <arpa/inet.h>

#include "scope.h"
#include "usbtmc.h"

#define USB_TIMEOUT 50000

#if BYTE_ORDER == LITTLE_ENDIAN
#define LE32(x) x
#elif BYTE_ORDER == BIG_ENDIAN
#define LE32(x) ((uint32_t)((((uint32_t)x)>>24) | ((((uint32_t)x)>>8) & 0xff00) | ((((uint32_t)x)<<8) & 0xff0000) | (((uint32_t)x)<<24)))
#else
#error BYTE_ORDER not defined/known!
#endif

/* TODO: fix memory leak here: */
#define USB_ERROR(s, x) do { if (x < 0) { fprintf(stderr, "usb %s: %s\n", s, usb_strerror()); usb_reset(sc->usb.dev); return 0; } } while(0)

/* This routine locates a scope by VID/PID and returns a struct scope* for it */
static struct scope* usbtmc_find_scope() {
	struct usb_bus *bus;
	struct usb_device *dev=NULL;
	struct usb_dev_handle *devh;

	struct scope *sc;

	usb_find_busses();
	usb_find_devices();

	for (bus=usb_busses; bus; bus=bus->next) {
		for (dev=bus->devices; dev; dev=dev->next) {
			if (dev->descriptor.idVendor == 0x400 && dev->descriptor.idProduct == 0x5dc) {
				devh = usb_open(dev);
				if (devh == NULL)
					return NULL;

				sc = calloc(1, sizeof(struct scope));
				if (sc == NULL) {
					perror("calloc");
					exit(EXIT_FAILURE);
				}

				sc->usb.dev = devh;

				/* TODO: FIXME */
				sc->usb.brokenRigol = 1;
				sc->usb.ep_bulk_out = 0x01;
				sc->usb.ep_bulk_in = 0x82;
				sc->usb.wMaxPacketSize_in = 0x40;
				
				return sc;
			}
		}
	}

	return NULL;
}

static unsigned char usbtmc_status(struct scope *sc)
{
	int r;
	unsigned char status[3];

	sc->usb.bTag++;

	r = usb_control_msg(sc->usb.dev, 0xA1,
		USB488_CTL_READ_STATUS_BYTE,
		(sc->usb.bTag & 0x7f), 0, (char*)status, 3,
		USB_TIMEOUT);

	if ((r != 3) || (status[0] != 0x01) || (status[1] != (sc->usb.bTag & 0x7f))) {
		printf("READ_STATUS_BYTE failed: %d 0x%x 0x%x 0x%x\n", r, status[0], status[1], status[2]);
		return 0xff;
	}

	return status[2];
}

static struct usbtmc_capabilities* usbtmc_get_capabilities(struct scope *sc)
{
	int r;
	static struct usbtmc_capabilities res;

	r = usb_control_msg(sc->usb.dev, 0xA1, 
		USBTMC_CTL_GET_CAPABILITIES,
		0, 0, (char*)&res, sizeof(struct usbtmc_capabilities),
		USB_TIMEOUT);
	if (r != sizeof(struct usbtmc_capabilities)) {
		printf("GET_CAPABILITIES failed: %s\n", usb_strerror());
		return NULL;
	}

	printf("USBTMC Version %x.%x Capabilities:\n", res.bcdUSBTMC[0], res.bcdUSBTMC[1]);
	if (res.USBTMCIFcapabilities & USBTMC_CAP_IF_INDICATOR_PULSE)
		printf("\tInterface supports indicator pulse\n");

	if (res.USBTMCIFcapabilities & USBTMC_CAP_IF_TALKONLY)
		printf("\tInterface is talk only\n");

	if (res.USBTMCIFcapabilities & USBTMC_CAP_IF_LISTENONLY)
		printf("\tInterface is listen only\n");

	if (res.USBTMCDEVcapabilities & USBTMC_CAP_DEV_TERMCHAR_SUPP)
		printf("\tDevice supports Termchar\n");

	printf("USB488 Version %x.%x Capabilities:\n", res.bcdUSB488[0], res.bcdUSB488[1]);

	if (res.USB488IFcapabilities & USB488_CAP_IF_4882)
		printf("\tInterface is 488.2 compliant\n");

	if (res.USB488IFcapabilities & USB488_CAP_IF_LOCKOUT)
		printf("\tInterface supports local lockout\n");

	if (res.USB488IFcapabilities & USB488_CAP_IF_TRIGGER)
		printf("\tInterface supports TRIGGER\n");

	if (res.USB488DEVcapabilities & USB488_CAP_DEV_SCPI)
		printf("\tDevice is SCPI compliant\n");

	if (res.USB488DEVcapabilities & USB488_CAP_DEV_SR1)
		printf("\tDevice is SR1 capable\n");

	if (res.USB488DEVcapabilities & USB488_CAP_DEV_RL1)
		printf("\tDevice is RL1 capable\n");

	if (res.USB488DEVcapabilities & USB488_CAP_DEV_DT1)
		printf("\tDevice is DT1 capable\n");

	return &res;
}

/*
 * Send a scpi-command to the scope. The response goes into the buffer
 * called resp, with a size of resplen. If resp==NULL, no response
 * is requested.
 */
int usbtmc_sendscpi(struct scope *sc, char* cmd, 
		unsigned char *resp, int resplen) {
	int len,r;
	int cmdlen = strlen(cmd);
	struct usbtmc_header *req;

	sc->usb.bTag++;

	len = sizeof(struct usbtmc_header) + cmdlen;
	if (len%4)
		len += 4 - (len%4);
	
	req = calloc(1, len);
	if (req == NULL) {
		perror("calloc");
		exit(EXIT_FAILURE);
	}

	req->MsgID = USBTMC_DEV_DEP_MSG_OUT;
	req->bTag = sc->usb.bTag;
	req->bTagInverse = ~sc->usb.bTag;
	req->TransferSize = LE32(cmdlen);
	req->bmTransferAttributes = USBTMC_TRANSFERATTRIB_EOM;
	memcpy(req->msg, cmd, cmdlen);

	if (sc->usb.brokenRigol) {
		r=usb_bulk_write(sc->usb.dev, sc->usb.ep_bulk_out,
			(char*)req, sizeof(struct usbtmc_header),
			USB_TIMEOUT);
		USB_ERROR("USBTMC_DEV_DEP_MSG_OUT1", r);

		r=usb_bulk_write(sc->usb.dev, sc->usb.ep_bulk_out,
			(char*)&(req->msg), len - sizeof(struct usbtmc_header),
			USB_TIMEOUT);
		USB_ERROR("USBTMC_DEV_DEP_MSG_OUT2", r);
	} else {
		r=usb_bulk_write(sc->usb.dev, sc->usb.ep_bulk_out,
			(char*)req, len, USB_TIMEOUT);
		USB_ERROR("USBTMC_DEV_DEP_MSG_OUT", r);
	}

	free(req);

	if (resp != NULL && resplen != 0) {
		unsigned char *buff;
		struct usbtmc_header *res;
		int bytes_read;

		sc->usb.bTag++;

		req = calloc(1, sizeof(struct usbtmc_header));
		if (req == NULL) {
			perror("calloc");
			exit(EXIT_FAILURE);
		}

		req->MsgID = USBTMC_REQUEST_DEV_DEP_MSG_IN;
		req->bTag = sc->usb.bTag;
		req->bTagInverse = ~sc->usb.bTag;
		req->TransferSize = LE32(sc->usb.wMaxPacketSize_in);
		req->bmTransferAttributes = 0;
		req->TermChar = 0;

		/* send read command */
		r=usb_bulk_write(sc->usb.dev, sc->usb.ep_bulk_out,
			(char*)req, sizeof(struct usbtmc_header), USB_TIMEOUT);
		USB_ERROR("USBTMC_REQUEST_DEV_DEP_MSG_IN", r);

		free(req);

		buff=malloc(sc->usb.wMaxPacketSize_in);
		if (buff == NULL) {
			perror("malloc");
			exit(EXIT_FAILURE);
		}

		r=usb_bulk_read(sc->usb.dev, sc->usb.ep_bulk_in,
			(char*)buff, sc->usb.wMaxPacketSize_in, USB_TIMEOUT);
		USB_ERROR("USBTMC_DEV_DEP_MSG_IN1", r);

		if (r < sizeof(struct usbtmc_header)) {
			fprintf(stderr, "Short read!\n");
			return 0;
		}

		bytes_read = r - sizeof(struct usbtmc_header);
		
		res = (struct usbtmc_header*)buff;
		len = LE32(res->TransferSize);

		memmove(buff, buff + sizeof(struct usbtmc_header), bytes_read);

		buff = realloc(buff, len);
		if (buff == NULL) {
			perror("realloc");
			exit(EXIT_FAILURE);
		}

		while ((len - bytes_read) > 0) {
			r=usb_bulk_read(sc->usb.dev, sc->usb.ep_bulk_in,
				(char*)buff + bytes_read, len - bytes_read,
				USB_TIMEOUT);
			USB_ERROR("USBTMC_DEV_DEP_MSG_INx", r);

			bytes_read += r;
		}

		/* TODO: FIXME */
		if (bytes_read > resplen) {
			fprintf(stderr, "Response buffer to small: %d instead of %d bytes!\n",
				resplen, bytes_read);
			bytes_read = resplen;
		}

		memcpy(resp, buff, bytes_read);
		free(buff);

		return bytes_read;
	}
	return 0;
}

void usbtmc_claim(struct scope *sc)
{
	usb_claim_interface(sc->usb.dev, 0);
}

void usbtmc_release(struct scope *sc)
{
	usb_release_interface(sc->usb.dev, 0);
}

/* Initialize the scope. */
struct scope* usbtmc_initscope(void) {
	int r;
	uint32_t vidpid;
	struct scope *sc;

	/* Init libusb */
	usb_init();
	/* Locate and open the scope */
	sc = usbtmc_find_scope();
	if (!sc) {
		return NULL;
	}

	usbtmc_claim(sc);
	sc->usb.cap = usbtmc_get_capabilities(sc);
	printf("Device status: 0x%x\n", usbtmc_status(sc));
	/* The following code isn't really necessary, the program works
	   OK without it too. */
	r=usb_control_msg(sc->usb.dev, 0xC8, 9, 0, 0, (char*)&vidpid, 4, USB_TIMEOUT);
	usbtmc_release(sc);
	if (r < 0) {
		fprintf (stderr, "Error %d sending init message: %s\n", 
				r, strerror (-r));
		fprintf (stderr, "Do you have permission on the USB device?\n");
		exit (1);
	}
	if (LE32(vidpid)!=0x40005dc) {
		fprintf(stderr,"Init: buff[%i]=%x\n",r,LE32(vidpid));
	}
	return sc;
}

void usbtmc_close(struct scope *sc)
{
	/* Free up and exit */
	usb_close(sc->usb.dev);
}
